#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import warnings
from loguru import logger as loguru_logger


class _Config(object):
    """Class for configuration of logger.
    """
    _LEVEL_DICT = {0: "ERROR", 1: "WARNING", 2: "INFO", 3: "DEBUG"}

    def __init__(self):
        self._logger = loguru_logger
        self._logger_level = 2

    @property
    def logger_level(self):
        """int: integer to indicate the logging level
        """
        return self._logger_level

    def logger(self, level):
        """Update configuration of logger.

        Args:
            level (int): log level (0: ERROR, 1: WARNING, 2: INFO, 3: DEBUG)
        """
        self._logger.remove()
        self._logger.add(
            sys.__stdout__,
            level=self._LEVEL_DICT[level],
            format="{time:YYYY-MM-DD at HH:mm:ss} | <level>{level}</level> | <level>{message}</level>",
        )
        self._logger_level = level

    def error(self, message):
        """Logging raised exception.

        Args:
            message (str): message to show
        """
        self._logger.error(message, exec_info=True)

    def warning(self, message, category=None):
        """Raise warning.

        Args:
            message (str): message to show
            category (Warning or None): category of warning or None (DeprecationWarning)
        """
        self._logger.warning(message)
        warnings.warn("deprecated callable was used", category, stacklevel=2)

    def info(self, message):
        """Show information.

        Args:
            message (str): message to show
        """
        self._logger.info(message)

    def debug(self, message):
        """Show debug message.

        Args:
            message (str): message to show
        """
        self._logger.debug(message)


config = _Config()
config.logger(level=2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pandas as pd
from covsirphy.util.term import Term
from covsirphy.downloading._db import _DataBase


class _WPP(_DataBase):
    """
    Access "World Population Prospects by United nations" server.
    https://population.un.org/wpp/
    """
    TOP_URL = "https://population.un.org/wpp/Download/Files/1_Indicators%20(Standard)/CSV_FILES/"
    # File title without extensions and suffix
    TITLE = "world-population-prospects"
    # Dictionary of column names
    COL_DICT = {
        "Time": "Year",
        "ISO3_code": Term.ISO3,
        "PopTotal": Term.N,
    }
    ALL_COLS = [Term.DATE, Term.ISO3, Term.PROVINCE, Term.CITY, Term.N]
    # Stdout when downloading (shown at most one time)
    STDOUT = "Retrieving datasets from World Population Prospects https://population.un.org/wpp/"
    # Citation
    CITATION = 'United Nations, Department of Economic and Social Affairs,' \
        ' Population Division (2022). World Population Prospects 2022, Online Edition.'

    def _country(self):
        """Returns country-level data.

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): observation date
                    - ISO3 (str): country names
                    - Province (object): NAs
                    - City (object): NAs
                    - Population (numpy.float64): population values
        """
        url = f"{self.TOP_URL}WPP2022_TotalPopulationBySex.zip"
        df = self._provide(url=url, suffix="_level1", columns=list(self.COL_DICT.keys()))
        df[self.DATE] = pd.to_datetime(df["Year"], format="%Y") + pd.offsets.DateOffset(months=6)
        df[self.PROVINCE] = self.NA
        df[self.CITY] = self.NA
        df[self.N] = df[self.N] * 1_000
        return df.dropna(how="any").loc[:, self.ALL_COLS]

    def _province(self, country):
        """Returns province-level data.

        Args:
            country (str): country name

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): observation date
                    - ISO3 (str): country names
                    - Province (object): NAs
                    - City (object): NAs
                    - Population (numpy.float64): population values
        """
        return pd.DataFrame(columns=self.ALL_COLS)

    def _city(self, country, province):
        """Returns city-level data.

        Args:
            country (str): country name
            province (str): province/state/prefecture name

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): observation date
                    - ISO3 (str): country names
                    - Province (object): NAs
                    - City (object): NAs
                    - Population (numpy.float64): population values
        """
        return pd.DataFrame(columns=self.ALL_COLS)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
import warnings
import geopandas as gpd
from unidecode import unidecode
from covsirphy.util.config import config
from covsirphy.util.validator import Validator
from covsirphy.util.filer import Filer
from covsirphy.util.term import Term


class _Geometry(Term):
    """Class to add geometry information to geo-spatial data.

    Args:
        data (pandas.DataFrame): geo-spatial data
            Index
                reset index
            Columns
                - column defined by @layer
                - the other columns
        layer (str): layer name which has location information
        directory (str): directory to save GeoJSON files
    """

    def __init__(self, data, layer, directory):
        self._df = Validator(data, f"{layer}-level data").dataframe(columns=[layer])
        self._layer = layer
        self._filer = Filer(directory=directory)

    def to_geopandas(self, iso3, natural_earth):
        """Add geometry information with GeoJSON file of "Natural Earth" GitHub repository to data.

        Args:
            iso3 (str or None): ISO3 code (for the province/city-level data) or None (country-level data)
            natural_earth (str or None): title of GeoJSON file (without extension) of "Natural Earth" GitHub repository or None (automatically determined)

        Returns:
            geopandas.GeoDataFrame:
                Index:
                    - reset index
                Columns:
                    - columns included in @data of _Geometry()
                    - geometry: geometric information

        Note:
            GeoJSON files are listed in https://github.com/nvkelso/natural-earth-vector/tree/master/geojson
            https://www.naturalearthdata.com/
            https://github.com/nvkelso/natural-earth-vector
            Natural Earth (Free vector and raster map data at naturalearthdata.com, Public Domain)
        """
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        gdf = self._natural_earth(title=natural_earth or self._natural_earth_parse_title(iso3=iso3))
        right_on = self.ISO3 if iso3 is None else "NAME"
        gdf = self._df.merge(gdf, how="left", left_on=self._layer, right_on=right_on)
        return gdf.rename(columns={f"{self.ISO3}_x": self.ISO3}).loc[:, [*self._df.columns.tolist(), "geometry"]]

    @staticmethod
    def _natural_earth_parse_title(iso3):
        """Find the best file title ("Natural Earth" GitHub repository) for the layer and country.

        Args:
            iso3 (str or None): ISO3 code (for the province/city-level data) or None (country-level data)

        Returns:
            str: parsed title
        """
        if iso3 is None:
            return "ne_110m_admin_0_countries"
        scale = "50m" if iso3 == "USA" else "10m"
        return f"ne_{scale}_admin_1_states_provinces"

    def _natural_earth(self, title):
        """Download GeoJSON files from "Natural Earth" GitHub repository.
        https://www.naturalearthdata.com/
        https://github.com/nvkelso/natural-earth-vector
        Natural Earth (Free vector and raster map data at naturalearthdata.com, Public Domain)

        Args:
            title (str): title of GeoJSON file (without extension) of "Natural Earth" GitHub repository

        Returns:
            geopandas.GeoDataFrame:
                Index:
                    - reset index
                Columns:
                    - ISO3: ISO3 codes
                    - NAME: Country names or province names
                    - geometry: geometric information

        Note:
            GeoJSON files are listed in https://github.com/nvkelso/natural-earth-vector/tree/master/geojson
        """
        file_dict = self._filer.geojson(title, driver="GeoJSON")
        filename = file_dict["filename"]
        if Path(filename).exists():
            gdf = gpd.read_file(filename)
        else:
            config.info("Retrieving GIS data from Natural Earth https://www.naturalearthdata.com/")
            url = f"https://raw.githubusercontent.com/nvkelso/natural-earth-vector/master/geojson/{title}.geojson"
            gdf = gpd.read_file(url)
            gdf.rename(columns={"ISO_A3": self.ISO3, "adm0_a3": self.ISO3, "name": "NAME"}, inplace=True)
            gdf["NAME"] = gdf["NAME"].fillna(self.NA).apply(unidecode)
            gdf["NAME"] = gdf["NAME"].replace(
                {"Xizang": "Tibet", "Inner Mongol": "Inner Mongolia", "S. Sudan": "South Sudan",
                 "Fr. S. Antarctic Lands": "French Southern and Antarctic Lands"})
            gdf.loc[gdf[self.ISO3] == "HKG", "NAME"] = "Hong Kong"
            gdf.loc[gdf[self.ISO3] == "-99", self.ISO3] = self._to_iso3(gdf.loc[gdf[self.ISO3] == "-99", "NAME"])
            gdf[self.ISO3] = gdf[self.ISO3].replace({"MAC": "CHN"})
            gdf.to_file(**file_dict)
        return gdf.reindex(columns=[self.ISO3, "NAME", "geometry"])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import warnings
from covsirphy.util.config import config
from covsirphy.util.error import NotRegisteredError, SubsetNotFoundError
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term
from covsirphy.gis.gis import GIS
from covsirphy.downloading._db_cs_japan import _CSJapan
from covsirphy.downloading._db_covid19dh import _COVID19dh
from covsirphy.downloading._db_google import _GoogleOpenData
from covsirphy.downloading._db_owid import _OWID
from covsirphy.downloading._db_wpp import _WPP


class DataDownloader(Term):
    """Class to download datasets from the recommended data servers.

    Args:
        directory (str or pathlib.Path): directory to save downloaded datasets
        update_interval (int): update interval of downloading dataset

    Note:
        Location layers are fixed to ["ISO3", "Province", "City"]
    """
    LAYERS = [Term.ISO3, Term.PROVINCE, Term.CITY]

    def __init__(self, directory="input", update_interval=12, **kwargs):
        self._directory = directory
        self._update_interval = Validator(update_interval, "update_interval").int(value_range=(0, None))
        if "verbose" in kwargs:
            verbose = kwargs.get("verbose", 2)
            config.logger(level=verbose)
            config.warning(
                f"Argument verbose was deprecated, please use covsirphy.config.logger(level={verbose}) instead.")
        self._gis = GIS(layers=self.LAYERS, country=self.ISO3, date=self.DATE)

    def layer(self, country=None, province=None, databases=None):
        """Return the data at the selected layer.

        Args:
            country (str or None): country name or None
            province (str or None): province/state/prefecture name or None
            databases (list[str] or None): databases to use or None (japan, covid19dh, google, owid).
                "japan": COVID-19 Dataset in Japan,
                "covid19dh": COVID-19 Data Hub,
                "google": COVID-19 Open Data by Google Cloud Platform (deprecated),
                "owid": Our World In Data,
                "wpp": World Population Prospects by United nations.

        Note:
            "google" for @database was deprecated and refer to https://github.com/lisphilar/covid19-sir/issues/1223

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    Date (pandas.Timestamp): observation date
                    ISO3 (str): country names
                    Province (str): province/state/prefecture names
                    City (str): city names
                    Country (str): country names (top level administration)
                    Province (str): province names (2nd level administration)
                    ISO3 (str): ISO3 codes
                    Confirmed (pandas.Int64): the number of confirmed cases
                    Fatal (pandas.Int64): the number of fatal cases
                    Recovered (pandas.Int64): the number of recovered cases
                    Population (pandas.Int64): population values
                    Tests (pandas.Int64): the number of tests
                    Product (pandas.Int64): vaccine product names
                    Vaccinations (pandas.Int64): cumulative number of vaccinations
                    Vaccinations_boosters (pandas.Int64): cumulative number of booster vaccinations
                    Vaccinated_once (pandas.Int64): cumulative number of people who received at least one vaccine dose
                    Vaccinated_full (pandas.Int64): cumulative number of people who received all doses prescribed by the protocol
                    School_closing
                    Workplace_closing
                    Cancel_events
                    Gatherings_restrictions
                    Transport_closing
                    Stay_home_restrictions
                    Internal_movement_restrictions
                    International_movement_restrictions
                    Information_campaigns
                    Testing_policy
                    Contact_tracing
                    Stringency_index
                    (deprecated) Mobility_grocery_and_pharmacy: % to baseline in visits (grocery markets, pharmacies etc.)
                    (deprecated) Mobility_parks: % to baseline in visits (parks etc.)
                    (deprecated) Mobility_transit_stations: % to baseline in visits (public transport hubs etc.)
                    (deprecated) Mobility_retail_and_recreation: % to baseline in visits (restaurant, museums etc.)
                    (deprecated) Mobility_residential: % to baseline in visits (places of residence)
                    (deprecated) Mobility_workplaces: % to baseline in visits (places of work)

        Note:
            When @country is None, country-level data will be returned.

        Note:
            When @country is a string and @province is None, province-level data in the country will be returned.

        Note:
            When @country and @province are strings, city-level data in the province will be returned.
        """
        db_dict = {
            "japan": _CSJapan,
            "covid19dh": _COVID19dh,
            "owid": _OWID,
            "wpp": _WPP,
            # Deprecated
            "google": _GoogleOpenData,
        }
        all_databases = ["japan", "covid19dh", "google", "owid"]  # "google" will be removed at 3.0.0
        selected = Validator(databases, "databases").sequence(default=all_databases, candidates=list(db_dict.keys()))
        self._gis = GIS(layers=self.LAYERS, country=self.ISO3, date=self.DATE)
        for database in selected:
            if database == "google":
                warnings.warn(
                    "Please use `databases=['japan', 'covid19dh', 'owid']` and refer to https://github.com/lisphilar/covid19-sir/issues/1223",
                    DeprecationWarning,
                    stacklevel=2
                )
            db = db_dict[database](
                directory=self._directory, update_interval=self._update_interval,)
            new_df = db.layer(country=country, province=province).convert_dtypes()
            if new_df.empty:
                continue
            self._gis.register(
                data=new_df, layers=self.LAYERS, date=self.DATE, citations=db.CITATION, convert_iso3=False)
        try:
            return self._gis.layer(geo=(country, province))
        except NotRegisteredError:
            raise SubsetNotFoundError(geo=(country, province)) from None

    def citations(self, variables=None):
        """
        Return citation list of the data sources.

        Args:
            variables (list[str] or None): list of variables to collect or None (all available variables)

        Returns:
            list[str]: citation list
        """
        return self._gis.citations(variables=variables)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import contextlib
from datetime import datetime, timezone, timedelta
from pathlib import Path
from urllib.parse import urlparse
import warnings
import numpy as np
import pandas as pd
from unidecode import unidecode
from covsirphy.util.config import config
from covsirphy.util.term import Term


class _DataProvider(Term):
    """Extract datasets and save it locally.

    Args:
        update_interval (int or None): update interval of downloading dataset or None (do not update if we have files in local)
        stdout: stdout when downloading if the logger level > info
    """

    def __init__(self, update_interval, stdout):
        self._update_interval = update_interval
        self._stdout = stdout

    def latest(self, filename, url, columns, date, date_format):
        """Provide the last dataset as a dataframe, downloading remote files or reading local files.

        Args:
            filename (str or pathlib.Path): filename to save/read the download
            url (str): URL of the dataset
            columns (list[str]): column names the dataset must have
            date (str or None): column name of date
            date_format (str): format of date column, like %Y-%m-%d

        Returns:
            pandas.DataFrame
        """
        if not self.download_necessity(filename):
            with contextlib.suppress(ValueError):
                return self.read_csv(filename, columns, date=date, date_format=date_format)
        config.info(self._stdout)
        df = self.read_csv(url, columns, date=date, date_format=date_format)
        df.to_csv(filename, index=False)
        return df

    @staticmethod
    def _last_updated_local(path):
        """
        Return the date last updated of local file/directory.

        Args:
            path (str or pathlib.Path): name of the file/directory

        Returns:
            datetime.datetime: time last updated (UTC)
        """
        m_time = Path(path).stat().st_mtime
        date = datetime.fromtimestamp(m_time)
        return date.astimezone(timezone.utc).replace(tzinfo=None)

    def download_necessity(self, filename):
        """
        Return whether we need to get the data from remote servers or not,
        comparing the last update of the files.

        Args:
            filename (str): filename of the local file

        Returns:
            bool: whether we need to get the data from remote servers or not

        Note:
            If the last updated date is unknown, returns True.
            If update interval was set as None, return False.
            If @update_interval (of _DataProvider) hours have passed and the remote file was updated, return True.
        """
        if not Path(filename).exists():
            return True
        if self._update_interval is None:
            return False
        date_local = self._last_updated_local(filename)
        time_limit = date_local + timedelta(hours=self._update_interval)
        return datetime.now() >= time_limit

    @staticmethod
    def read_csv(path, columns, date, date_format):
        """Read the CSV file and return as a dataframe.

        Args:
            columns (list[str] or None): column names the dataset must have
            date (str or None): column name of date
            date_format (str): format of date column, like %Y-%m-%d

        Returns:
            pandas.DataFrame: downloaded data
        """
        warnings.filterwarnings("ignore", category=pd.errors.DtypeWarning)
        kwargs = {
            "header": 0, "usecols": columns, "encoding": "utf-8", "engine": "pyarrow",
            "parse_dates": None if date is None else [date], "date_parser": lambda x: datetime.strptime(x, date_format)
        }
        if urlparse(path).scheme:
            kwargs["storage_options"] = {"User-Agent": "Mozilla/5.0"}
        df = pd.read_csv(path, **kwargs)
        for col in df:
            with contextlib.suppress(TypeError):
                df[col] = df[col].apply(lambda x: unidecode(x) if len(x) else np.nan)
        return df
